'''
Handles night cycle for HVAC system
'''

def availability_managers(building,T_zone,T_set,dt,fan_avail):
    override = False
    for i in range(len(building.hvac['managers']['name'])):
        if building.hvac['managers']['type'][i]=='NightCycle':
            l = building.hvac['managers']['loop'][i]
            z_i = [z for z,x in enumerate(building.zones['air_loop']) if x==l] #zones on this loop
            
            if 'runtime_control_type' in building.hvac['managers'] and building.hvac['managers']['runtime_control_type'][i]=='fixedruntime':
                uncomfortable = False
                for x in z_i:
                    if ((T_zone[x]>(T_set['cool'][x] + building.hvac['managers']['thermostat_tolerance'][i])
                        or T_zone[x]<(T_set['heat'][x] - building.hvac['managers']['thermostat_tolerance'][i]))
                        and T_set['no_hvac'][x]==False):
                        uncomfortable = True
                        break
                if uncomfortable: ## outside bounds, need to override
                    on_loop = [i for i,x in enumerate(building.hvac['components']['loop']) if x==l]
                    for j in on_loop:
                        if building.hvac['components']['type'][j][:4]=='fan:' :
                            if fan_avail[building.hvac['components']['name'][j]] == 0:
                                fan_avail[building.hvac['components']['name'][j]] = min(1,building.hvac['managers']['runtime'][i]/dt)
                                override = True #turn fan on when system is off
                        elif building.hvac['components']['type'][j]=='airloophvac:unitaryheatcool':
                            u = building.unitary_heat_cool['name'].index(building.hvac['components']['name'][j])
                            if fan_avail[building.unitary_heat_cool['fan_name'][u]] == 0:
                                fan_avail[building.unitary_heat_cool['fan_name'][u]]= min(1,building.hvac['managers']['runtime'][i]/dt)
                                override = True #turn fan on when system is off
            else:
                print('need runtime control other than fixedruntime for night manager')

        elif building.hvac['managers']['type'][i]=='Scheduled':
            pass
        else:
            print('need to code other types of availability managers')
    return fan_avail,override


'''
Iterate to find central air and direct zone air temperature and flow that achieves temperature constraints
'''

from eagers.building.basic.psychometric import psychometric
from eagers.building.process.window_temperature import window_temperature
from eagers.building.basic.natural_convection import natural_convection
from eagers.building.components.water_coil import water_coil
from eagers.building.components.fan_calc import fan_calc
from eagers.building.components.manager_override import manager_override
from eagers.building.process.update_model_states_implicit import update_model_states_implicit
from eagers.building.components.availability_managers import availability_managers
from eagers.building.basic.leward_calc import leward_calc
from eagers.building.basic.wind_speed_calc import wind_speed_calc

def ideal_hvac(building,profile,leward,wind_speed,S,air_nodes,dry_bulb,mixing,infiltration,gains,T,w,T_set,w_set,occupancy,dt,des_day):
    air_density = 1.204 #kg/m^3  #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    n_z = len(T['zone'])
    n_sur = len(building.surfaces['name'])
    n_win = len(building.windows['name'])
    fresh_air = {}
    fresh_air['Tdb'] = dry_bulb
    fresh_air['w'] = w['air']+0
    fresh_air['h'] = psychometric(fresh_air,'h')
    cp_zone = [(1006 + 1860*w['zone'][i])*air_density for i in range(n_z)] #Specific heat in J/m^3
    T['zone_est'] = [j for j in T['zone']]
    T['surf_est'] = [j for j in T['surf']]
    w['zone_est'] = [j for j in w['zone']] #estimated values at the end of the time step
    T['central'] = []
    fresh_air_zone_req = design_zone_outdoor_air(building,occupancy) #requested fresh air of the zone
    net_H2O_gain = [gains['zone_latent'][i]/2264705 + air_density*(sum([mixing[i][j]*w['zone'][j] for j in range(len(mixing[i]))]) + infiltration[i]*w['air']) for i in range(len(infiltration))] #flows in m^3/s converted to kg/s, everything into kg/s of water 
    fan_avail = {}
    for i in building.fans:
        fan_avail[i] = profile[building.fans[i]['schedule']]
    
    Q_transient = [[],[]]
    Q_transient[0] = [(T['zone'][k] - T_set['cool'][k])*cp_zone[k]*building.zones['volume'][k]/dt for k in range(n_z)]
    if T_set['dual_sp']:
        Q_transient[1] = [(T['zone'][k] - T_set['heat'][k])*cp_zone[k]*building.zones['volume'][k]/dt for k in range(n_z)] #zone heat capacitance*deltaT/dt = power (W)

    Q_Tset = [[],[]]
    T_loop = [13 for i in range(len(building.hvac['loop']['name']))]
    w['loop'] = [.01 for i in range(len(building.hvac['loop']['name']))]
    window_wind_speed = [wind_speed[building.windows['ext_surf'][i]] for i in range(n_win)]
    window_leward = [leward[building.windows['ext_surf'][i]] for i in range(n_win)]
    window_air_temp = [dry_bulb - 0.0065*building.windows['height'][i]+273 for i in range(n_win)]

    tol = building.site['temp_tol']
    max_e = 10*tol
    count = 0
    while count<8 and max_e>0.1*tol:
        ## Windows
        T_interior = [T['surf_est'][i] for i in building.ctf['sur_state1']]
        T_int_K = [T_interior[i]+273 for i in range(len(T_interior))]
        window_zone_temp = [T['zone_est'][building.windows['zone_index'][i]]+273 for i in range(n_win)]        
        cp_window = [cp_zone[building.windows['zone_index'][i]] for i in range(n_win)]
        T['windows'],T['windows_int'],Q_windows,_ = window_temperature(building,window_wind_speed,window_leward,S,T['windows'],T['windows_int'],
                                        T['sky']+273,window_zone_temp,window_air_temp,T_int_K,cp_window)
        
        ## Predicts the zone energy exchange that achieves the temperature setpoints 
        ## Assume convection and mixing at expected zone/surface temperatures
        dT = [T_interior[i] - T['zone_est'][building.surfaces['zone_index'][i]] for i in range(n_sur)] #interior surface convection = h*A*(Tsur - Tzone) 
        h_interior = natural_convection(building.surfaces['normal'],dT,T_set['no_hvac'],building.convection['interior'])
        Q_mixing = [cp_zone[i]*(sum([mixing[i][j]*T['zone_est'][j] for j in range(n_z)]) - sum(mixing[i])*T['zone_est'][i])for i in range(n_z)]  #heat into zone via mixing from another zone
        Q_infiltration = [cp_zone[i]*infiltration[i]*(T['air_zone'][i] - T['zone_est'][i]) for i in range(n_z)] #heat into zone from air infiltration 
        Q_Tset = qset(T_set,building,gains['zone_sensible'],Q_windows,Q_transient,Q_infiltration,Q_mixing,T_interior,h_interior)#Energy required by the zone in W 
        
        ## determine the supply flow rate and temperature to each zone/loop that will result in the desired zone temperature (passive control within deadband)
        T_min,T_max = zone_limits(building,T['zone_est'])
        exhaust_flow = exhaust_fan_flow(building,profile,mixing)
        flow_imbalance = [infiltration[i] - exhaust_flow[i] for i in range(n_z)] #volumetric flow (m^3/s)

        if not des_day:
            central_flow = terminal_flow(building,Q_Tset,T_set,cp_zone,T_min,T_max,T['central'],flow_imbalance,fan_avail,T_loop,w['loop'],T['air_zone'][0])
            loop_flow = supply_flow_limit(building,profile,central_flow,fan_avail) #enforces minimum and maximum flow rates     
            loop_2_zone,central_flow = divide_loop_flow(building,loop_flow,central_flow)
            direct_flow = zone_hvac_flow(building,profile,Q_Tset)
            fresh_air_loop = loop_outdoor_air(building,profile,fresh_air_zone_req,central_flow,loop_flow,flow_imbalance)
            direct_fresh_air = direct_zone_outdoor_air(building,profile,Q_Tset,flow_imbalance,central_flow)
        else:
            central_flow,direct_flow,loop_flow,direct_fresh_air,fresh_air_loop = des_day_zone_flow(building,fresh_air_zone_req,cp_zone,Q_Tset,T_set,T_min,T_max,flow_imbalance)
        
        flow_exiting_zone,return_flow,plenum = plenum_return_flow(building,profile,central_flow,direct_flow,mixing,infiltration)
        direct_return_air,loop_return_air = return_air_calc(building,gains,T['zone_est'],w['zone_est'],return_flow,direct_flow,direct_fresh_air)
        T_loop,w['loop'],mixed_air_loop,air_nodes = loop_supply_setpoint(building,profile,air_nodes,fresh_air,fresh_air_loop,loop_return_air,loop_flow,central_flow,loop_2_zone,cp_zone,flow_exiting_zone,net_H2O_gain,T_set,w_set,Q_Tset,T_min,T_max,fan_avail,des_day)
        supply_flow,T,w = zones_supply_setpoint(building,T,w,central_flow,cp_zone,T_loop,direct_return_air,T_set,Q_Tset,T_min,T_max,fresh_air,direct_fresh_air,direct_flow,fan_avail)

        ## Estimate of zone conditions if HVAC logic is applied
        T_z_new,T['surf_est'],w['zone_est'],e_dict = update_model_states_implicit(building,T,S,wind_speed,leward,gains,dt,w,mixing,plenum,infiltration,supply_flow)

        ## add night manager here, if it turns things on or off, don't update T.zone_est, and don't count up
        fan_avail,override = availability_managers(building,T_z_new,T_set,dt,fan_avail)
        if override:
            max_e = 10*tol
        else:
            max_e = max([abs(T_z_new[i] - T['zone_est'][i]) for i in range(n_z)])
            cp_zone = [(1006 + 1860*w['zone_est'][i])*air_density for i in range(n_z)] #Specific heat in J/m^3
            T['zone_est'] = [j for j in T_z_new]
            count +=1    

    # print('ideal_hvac loops:  ' + str(count))
    # if any([T['zone_est'][j]<T_set['heat'][j]- building.hvac['managers']['thermostat_tolerance'][building.zones['air_loop'][j]] and not T_set['no_hvac'][j] for j in range(len(T['zone_est']))]):
    #     ## estimate shortfall of HVAC treatment
    #     zone_cool = [supply_flow['flow'][z]*cp_zone[z]*(supply_flow['Tdb'][z]- T_set['cool'][z]) for z in range(n_z)]
    #     zone_heat = [supply_flow['flow'][z]*cp_zone[z]*(supply_flow['Tdb'][z]- T_set['heat'][z]) for z in range(n_z)]
    #     cooling_shortfall_perc = [0 if zone_cool[z] == 0 else 100*max(0,-Q_Tset[0][z] - zone_cool[z])/(-Q_Tset[0][z]) for z in range(n_z)]
    #     heating_shortfall_perc = [0 if zone_heat[z] == 0 else 100*max(0,Q_Tset[1][z] - zone_heat[z])/Q_Tset[1][z] for z in range(n_z)]
    #     print('investigate why building not reaching setpoint')
    return air_nodes,central_flow,direct_flow,mixed_air_loop,direct_return_air,plenum,T,w,fan_avail

def qset(T_set,building,sensible_gain,Q_windows,Q_trans,Q_infiltration,Q_mixing,T_interior,h_interior):
    # Detrimes the energy requirment to be within acceptable temperature window
    # + is heating, - is cooling
    # if no heating is required to reach heating setpoint and no cooling required to reach cooling setpoint then temperature can drift naturally
    # Q_set[0][i] represents the heating required to reach the cooling setpoint in zone 1. A negative value requires cooling
    # Q_set[1][i] represents the heating required to reach the heating setpoint in zone 1. A positive value requires heating
    Q_set = [[],[]]
    n_z = len(T_set['cool'])
    n_sur = len(building.surfaces['name'])
    Q_surfaces = [sum([-h_interior[j]*(T_set['cool'][k]- T_interior[j])*building.surfaces['area'][j] for j in range(n_sur) if building.surfaces['zone_index'][j] ==k]) for k in range(n_z)] #estimated heat transfer to zone from walls/floor/ceiling 
    Q_set[0] = [-(sensible_gain[j] + Q_windows[j] + Q_mixing[j] + Q_infiltration[j] + Q_trans[0][j] + Q_surfaces[j]) if not T_set['no_hvac'][j] else 0 for j in range(n_z)]#Energy provided to the zone in W to achieve T_set.cool, + is heating
    if T_set['dual_sp']:
        Q_surfaces = [sum([-h_interior[j]*(T_set['heat'][k]- T_interior[j])*building.surfaces['area'][j] for j in range(n_sur) if building.surfaces['zone_index'][j] ==k]) for k in range(n_z)] #estimated heat transfer to zone from walls/floor/ceiling 
        Q_set[1] = [-(sensible_gain[j] + Q_windows[j] + Q_mixing[j] + Q_infiltration[j] + Q_trans[1][j] + Q_surfaces[j]) if not T_set['no_hvac'][j] else 0 for j in range(n_z)]#Energy provided to the zone in W to achieve T_set.heat, + is heating
    else:
        Q_set[1] = Q_set[0]
    return Q_set

def design_zone_outdoor_air(building,occupancy):
    '''
    ##Determine max fresh air requirements
    '''
    n_z = len(building.zones['name'])
    fresh_air_zone = [0 for i in range(n_z)]
    for i in range(n_z):
        try:
            z = building.setpoints['zone'][i]
            if building.setpoints['outdoor_flow_method'][i]=='flow/person':
                    fresh_air_zone[z] = building.setpoints['outdoor_flow_value'][i]*occupancy[z]
            elif building.setpoints['outdoor_flow_method'][i]=='flow/area':
                    fresh_air_zone[z] = building.setpoints['outdoor_flow_value'][i]*building.zones['floor_area'][z]
            elif building.setpoints['outdoor_flow_method'][i]=='flow/zone':
                    fresh_air_zone[z] = building.setpoints['outdoor_flow_value'][i]    
        except IndexError:
            pass
    return fresh_air_zone

def zone_limits(building,T_zone):
    T_min = [0 for i in range(len(T_zone))]
    T_max = [0 for i in range(len(T_zone))]
    for i in range(len(building.setpoints['name'])):
        z = building.setpoints['zone'][i]
        if building.setpoints['cooling_T_method'][i]=='supplyairtemperature':
            T_min[z] = building.setpoints['cooling_T_des'][i]
        else:
            T_min[z] -= building.setpoints['cooling_dT_des'][i]
        if building.setpoints['heating_T_method'][i]== 'supplyairtemperature':
            T_max[z] = building.setpoints['heating_T_des'][i]
        else:
            T_max[z] += building.setpoints['heating_dT_des'][i]
    for i in range(len(building.manager['name'])):
        if building.manager['type'][i]=='SingleZone:Reheat':
            z = building.zones['name'].index(building.manager['zone'][i])
            T_min[z] = building.manager['min_temp'][i]
            T_max[z] = building.manager['max_temp'][i]
    return T_min,T_max

def  exhaust_fan_flow(building,profile,mixing):
    ##exhaust fans
    n_z = len(building.zones['name'])
    exhaust_flow = [-sum([mixing[j][i] for j in range(n_z)]) for i in range(n_z)]#balanced exhaust flow (m^3/s)
    for i in building.fans:
        if building.fans[i]['exhaust']:
            exhaust_flow[building.fans[i]['exhaust_zone']] += building.fans[i]['flow_rate']*profile[building.fans[i]['schedule']] #volumetric flow (m^3/s)   #See page 1601 of the reference manual
    return exhaust_flow

def terminal_flow(building,Q_Tset,T_set,cp_zone,T_min,T_max,T_central,flow_imbalance,fan_avail,T_loop,w_loop,T_air):
    ##do VAV terminal air calculations  
    #Determine flow rate to zones, based on fan flow limits, fresh air requirements and if HVAC is on
    air_density = 1.204 #kg/m^3  #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    hv = building.hvac
    n_z = len(T_min)
    zone_flow_req = [0 for i in range(n_z)]
    for i in range(n_z):
        if T_set['no_hvac'][i]:
            pass
        elif Q_Tset[1][i]>0: #heating
            zone_flow_req[i] = Q_Tset[1][i]/(cp_zone[i]*(T_max[i] - T_set['heat'][i]))
        elif Q_Tset[0][i]<0:#cooling
            zone_flow_req[i] =  Q_Tset[0][i]/(cp_zone[i]*(T_min[i] - T_set['cool'][i]))

    n = len(hv['terminals']['name'])
    term_flow = [0 for i in range(n)]
    min_term_flow = [0 for i in range(n)]
    if 'min_flow' in hv['terminals']:
        for i in range(n):
            if hv['terminals']['min_flow'][i] == None:
                min_term_flow[i] = 0
            else:
                min_term_flow[i] = hv['terminals']['min_flow'][i]/building.zones['multiplier'][hv['terminals']['zone'][i]]        
    max_term_flow = [hv['terminals']['max_flow'][i]/building.zones['multiplier'][hv['terminals']['zone'][i]] for i in range(n)]

    for i in range(len(hv['loop']['name'])):# find the fan on the air loop and adjust min/max flow by fan availability
        term_i = []
        z_i = []
        m = []
        no_flow = False
        for x in range(len(building.air_supply_equip['loop'])):
            if building.air_supply_equip['loop'][x] ==i and (building.air_supply_equip['type'][x][:4]=='fan:'):
                f = building.air_supply_equip['name'][x]
        for j , k in enumerate( hv['terminals']['loop']):
            if k==i:
                min_term_flow[j] = min_term_flow[j]*fan_avail[f]
                max_term_flow[j] = max_term_flow[j]*fan_avail[f]
                if fan_avail[f]==0:
                    no_flow = True
        
                #Cooling mode:supply temperature is based on limiting zone, other zones operate with reduced flow. If a terminal minimum flow is hit, re-heat prevents from overcooling
                #Heating mode: minimum flow. re_heat does all necessary heating
                term_i.append(j)
                z_i.append(hv['terminals']['zone'][j])
                m.append(building.zones['multiplier'][hv['terminals']['zone'][j]])
        if not no_flow:
            if hv['terminals']['type'][term_i[0]]=='SingleDuct:Uncontrolled': #what if terminals on same loop are of different variety? Not an issue with default buildings
                for j in range(len(z_i)):
                    term_flow[term_i[j]] =  zone_flow_req[z_i[j]] #volumetric flow m^3/s
            elif hv['terminals']['type'][term_i[0]]=='SingleDuct:ConstantVolume:Reheat':
                    print('has constantvolume:reheat, dont know what to do')
            elif hv['terminals']['type'][term_i[0]]=='SingleDuct:VAV:Reheat':
                if len(z_i) == 1:#loop serves a single zone, heating = min flow, vary temp ... cooling = supply temp, vary flow
                    if Q_Tset[0][z_i[0]]<0 and T_set['no_hvac'][z_i[0]]==False: #cooling
                        term_flow[term_i[0]] = zone_flow_req[z_i[0]]
                    else: #if not cooling, set to minimuim (bottom of page 694)
                        term_flow[term_i[0]] = min_term_flow[term_i[0]]
                else: #loop serves multiple zones, must supply at temperature that provides sufficient cooling to all zones at max flow
                    #find the limiting zone, i.e. max flow at this supply temperature = perfect cooling. Other zones have re-heat
                    if any([Q_Tset[0][x]<0 and T_set['no_hvac'][x]==False for x in z_i]): #loop in cooling mode for at least 1 zone
                        cool_T = [max(T_min[z_i[j]],T_set['cool'][z_i[j]] + Q_Tset[0][z_i[j]]/(cp_zone[z_i[j]]*max_term_flow[term_i[j]])) for j in range(len(z_i))]
                        min_cooling_T = min(cool_T)
                        if len(T_central)>0:
                            min_cooling_T = min(min_cooling_T,min([T_central[z_k] for z_k in z_i]))
                        ## If it can reach below min_cooling without chillers, then preferable to be at lower temperature and reduced flow. 
                        if T_air<min_cooling_T:
                            min_cooling_T = max(max([T_min[zk] for zk in z_i]),T_air)

                        for j in range(len(z_i)):
                            no_treat_cooling = min_term_flow[term_i[j]]*cp_zone[z_i[j]]*(T_set['heat'][z_i[j]] - min_cooling_T) #m*cp*deltaT at minimum flow and temperature
                            if Q_Tset[0][z_i[j]]<0 and T_set['no_hvac'][z_i[j]]==False:
                                term_flow[term_i[j]] = Q_Tset[0][z_i[j]]/(cp_zone[z_i[j]]*(min_cooling_T - T_set['cool'][z_i[j]])) #volumetric flow m^3/s
                            elif (Q_Tset[1][z_i[j]] + no_treat_cooling)>0 and T_set['no_hvac'][z_i[j]] == False:
                                term_flow[term_i[j]] = (Q_Tset[1][z_i[j]] + no_treat_cooling)/(cp_zone[z_i[j]]*(T_max[z_i[j]] - T_set['heat'][z_i[j]])) #volumetric flow m^3/s
                            else: 
                                term_flow[term_i[j]] = 0
                    else: #passive or heating
                        for j in range(len(z_i)):
                            if Q_Tset[1][z_i[j]]>0:
                                if hv['terminals']['reheat_coil_type'][term_i[j]]=='coil:heating:water':
                                    name = hv['terminals']['reheat_coil_name'][term_i[j]]
                                    max_flow = air_density*building.hvac['terminals']['max_flow'][term_i[j]]
                                    air,_,_,_ = water_coil(building,name,T_set['heat'][z_i[j]],Q_Tset[1][z_i[j]]*m[j],[T_loop[i],w_loop[i],max_flow])
                                    term_flow[term_i[j]] = air['m_dot']/air_density/m[j]
                                else:
                                    term_flow[term_i[j]] = Q_Tset[1][z_i[j]]/(cp_zone[z_i[j]]*(T_max[z_i[j]]-T_set['heat'][z_i[j]]))
            else:
                print('need additional air terminal types')
    central_flow = [0 for i in range(n_z)]
    for i in range(len(hv['terminals']['name'])):
        if T_set['no_hvac'][hv['terminals']['zone'][i]]==False:
            term_flow[i] =  max(term_flow[i], -flow_imbalance[hv['terminals']['zone'][i]]) #volumetric flow m^3/s
        term_flow[i] =  min(max(term_flow[i],min_term_flow[i]),max_term_flow[i])
        central_flow[hv['terminals']['zone'][i]] = term_flow[i]
    return central_flow

def supply_flow_limit(building,profile,central_flow,fan_avail):
    #reduce flow if any equipment in air_loop limits flow
    n_z = len(building.zones['name'])
    n_l = len(building.hvac['loop']['name'])
    limit_flow = [x for x in building.hvac['loop']['max_flow']] #volumetric flow (m3/s)
    minimum_flow = [0 for x in building.hvac['loop']['max_flow']] #volumetric flow (m3/s)
    ##go through equipment and reduce if schedule is reduced
    for i in range(len(building.air_supply_equip['name'])):
        loop = building.air_supply_equip['loop'][i]
        if building.air_supply_equip['type'][i]=='CoilSystem:Cooling:DX' or building.air_supply_equip['type'][i]=='Coil:Cooling:Water':
                k = building.coils_cooling['name'].index(building.air_supply_equip['name'][i])
                sched = building.coils_cooling['schedule'][k]
                limit_flow[loop] = min(limit_flow[loop],building.coils_cooling['rated_air_flow'][k]*profile[sched]) # #mass flow (kg/s)
        elif building.air_supply_equip['type'][i][:4]=='fan:':
                fan_name = building.air_supply_equip['name'][i]
                minimum_flow[loop],limit_flow[loop] = fan_flow_limit(building,fan_avail[fan_name],minimum_flow[loop],limit_flow[loop],fan_name)
        elif building.air_supply_equip['type'][i]=='airloophvac:unitaryheatcool':
                u = building.unitary_heat_cool['name'].index(building.air_supply_equip['name'][i])
                fan_name = building.unitary_heat_cool['fan_name'][u]
                minimum_flow[loop],limit_flow[loop] = fan_flow_limit(building,fan_avail[fan_name],minimum_flow[loop],limit_flow[loop],fan_name)         
    loop_flow = [0 for i in range(n_l)]
    for i in range(n_z):
        if building.zones['air_loop'][i]!=None:
            loop_flow[building.zones['air_loop'][i]] += central_flow[i]*building.zones['multiplier'][i] #volumetric flow (m3/s)
    for i in range(n_l):
        if loop_flow[i]>limit_flow[i]:
            loop_flow[i]=limit_flow[i]
        elif loop_flow[i]<minimum_flow[i]:
            loop_flow[i]=minimum_flow[i]
    return loop_flow

def fan_flow_limit(building,fan_avail,minimum,limit,f):
    if building.fans[f]['type']=='ConstantVolume' or building.fans[f]['type']=='OnOff':
        limit_flow = building.fans[f]['flow_rate']*fan_avail #volumetric flow (m3/s)
        min_flow = building.fans[f]['flow_rate']*fan_avail #volumetric flow (m3/s)
    elif building.fans[f]['type']=='VariableVolume':
        limit_flow = min(limit,building.fans[f]['flow_rate']*fan_avail) #volumetric flow (m3/s)
        min_flow = max(minimum,building.fans[f]['min_flow_rate']*fan_avail) #volumetric flow (m3/s)   
    elif building.fans[f]['type']=='Fan:SystemModel':
            print('need to program Fan:SystemModel')
    return min_flow,limit_flow

def divide_loop_flow(building,loop_flow,central_flow):
    n_z = len(building.zones['name'])
    n_l = len(building.hvac['loop']['name'])
    loop_2_zone = [[0 for j in range(n_l)] for i in range(n_z)]
    zone_request= [[0 for j in range(n_l)] for i in range(n_z)]
    loop_request = [0 for i in range(n_l)]
    adj_central_flow = []
    for i in range(n_z):
        l=building.zones['air_loop'][i]
        if l!=None:
            zone_request[i][l]= central_flow[i]*building.zones['multiplier'][i]
            loop_request[l] += zone_request[i][l]
            
    for l in range(n_l):
        if loop_flow[l]>0 and loop_request[l]==0: #split flow based on heating/cooling request flow #avoid NaN, split to zones based on max zone flow
            n_t = len(building.hvac['terminals']['name'])
            for j in range(n_t):
                if building.zones['air_loop'][building.hvac['terminals']['zone'][j]]==l:
                    zone_request[building.hvac['terminals']['zone'][j]][l] += building.hvac['terminals']['max_flow'][j]
            loop_request[l] = sum([zone_request[i][l] for i in range(n_z)])
        for i in range(n_z):
            if loop_request[l]>0:
                loop_2_zone[i][l] = zone_request[i][l]/loop_request[l]
    for i in range(n_z):
        adj_central_flow.append(sum([loop_flow[l]*loop_2_zone[i][l] for l in range(n_l)])/building.zones['multiplier'][i]) #loop flow adjusted in air loop flow constraints were active
    return loop_2_zone,adj_central_flow

def zone_hvac_flow(building,profile,Q_Tset):
    #flow for zones not on a HVAC loop (unit heaters)
    n_z = len(building.zones['name'])
    direct_flow = [0 for i in range(n_z)]
    for i in building.unitary_sys:
        f = building.unitary_sys[i]['fan_name']
        z = building.unitary_sys[i]['zone']
        if building.unitary_sys[i]['type']=='UnitHeater':
            direct_flow[z] = building.fans[f]['flow_rate']*profile[building.fans[f]['schedule']]/building.zones['multiplier'][z]
        elif building.unitary_sys[i]['type']=='FourPipeFanCoil':
            direct_flow[z] = building.fans[f]['flow_rate']*profile[building.fans[f]['schedule']]/building.zones['multiplier'][z]
        elif building.unitary_sys[i]['type']== 'PackagedTerminalAirConditioner':
                ## need to find part-load-ratio (PLR) see page 1541
                ## later infer PLR from this flow
                if Q_Tset[1][z]>0: #heating
                    direct_flow[z] = building.unitary_sys[i]['heating_air_flow']/building.zones['multiplier'][z]
                elif Q_Tset[0][z]<0: #cooling
                    direct_flow[z] = building.unitary_sys[i]['cooling_air_flow']/building.zones['multiplier'][z]
                else:
                    direct_flow[z] = building.unitary_sys[i]['no_load_air_flow']/building.zones['multiplier'][z]
    return direct_flow

def loop_outdoor_air(building,profile,fresh_air_zone_req,central_flow,loop_flow,flow_imbalance):
    #fresh air specified by a AirLoop HVAC
    nl = len(building.hvac['loop']['name'])
    n_z = len(building.zones['name'])
    fresh_air_loop = [0 for i in range(nl)]
    for i in range(len(building.hvac['outdoor_air']['name'])): #loop through outside_air controllers to get minimum
        fresh_air_loop[building.hvac['outdoor_air']['loop'][i]] = building.hvac['outdoor_air']['min_flow'][i]*profile[building.hvac['outdoor_air']['min_air_schedule'][i]] 
    for l in range(nl):
        flow_imbal = sum([-flow_imbalance[i]*building.zones['multiplier'][i] for i in range(n_z) if building.zones['air_loop'][i]==l])
        fresh_air_loop[l] = max(fresh_air_loop[l],flow_imbal) #mass flow that must be made up by fresh air(kg/s)
    return fresh_air_loop

def direct_zone_outdoor_air(building,profile,Q_Tset,flow_imbalance,central_flow):
    #fresh air specified by ZoneHVAC units
    n_z = len(building.zones['name'])
    direct_fresh_air = [0 for i in range(n_z)]
    for i in building.unitary_sys:
        z = building.unitary_sys[i]['zone']
        if building.unitary_sys[i]['type']=='FourPipeFanCoil':
            if ('outdoor_air_schedule' in building.unitary_sys[i] and building.unitary_sys[i]['outdoor_air_schedule'] != None):
                direct_fresh_air[z] = building.unitary_sys[i]['outdoor_air_flow']*profile[building.unitary_sys[i]['outdoor_air_schedule']]/building.zones['multiplier'][z]
            else:
                direct_fresh_air[z] = building.unitary_sys[i]['outdoor_air_flow']/building.zones['multiplier'][z]
        elif building.unitary_sys[i]['type']=='PackagedTerminalAirConditioner':
                if Q_Tset[1][z]>0: #heating
                    direct_fresh_air[z] = building.unitary_sys[i]['heating_outdoor_air_flow']*profile[building.unitary_sys[i]['fan_schedule']]/building.zones['multiplier'][z]
                elif Q_Tset[0][z]<0: #cooling
                    direct_fresh_air[z] = building.unitary_sys[i]['cooling_outdoor_air_flow']*profile[building.unitary_sys[i]['fan_schedule']]/building.zones['multiplier'][z]
                else:
                    direct_fresh_air[z] = building.unitary_sys[i]['no_load_outdoor_air_flow']*profile[building.unitary_sys[i]['fan_schedule']]/building.zones['multiplier'][z]
    d2 = [-(flow_imbalance[i] + central_flow[i]) for i in range(n_z)]
    direct_fresh_air = [max(direct_fresh_air[i],d2[i]) for i in range(n_z)]
    return direct_fresh_air


def plenum_return_flow(building,profile,central_flow,direct_flow,mixing,infiltration):
    ##exhaust fans
    n_z = len(building.zones['name'])
    n_p = len(building.hvac['plenum']['name'])
    exhaust_flow = [-sum([mixing[j][i] for j in range(n_z)]) for i in range(n_z)]#balanced exhaust flow (m^3/s)
    for i in building.fans:
        if building.fans[i]['exhaust']:
            exhaust_flow[building.fans[i]['exhaust_zone']] += building.fans[i]['flow_rate']*profile[building.fans[i]['schedule']] #volumetric flow (m^3/s)   #See page 1601 of the reference manual
    hvac_supply = [central_flow[i] + direct_flow[i] for i in range(n_z)]
    for i in building.unitary_sys:
        if 'terminal' in building.unitary_sys[i]:
            z = building.unitary_sys[i]['zone']
            hvac_supply[z] = direct_flow[z]

    return_flow = [hvac_supply[i] + sum(mixing[i]) - sum([mixing[j][i] for j in range(n_z)]) + infiltration[i] - exhaust_flow[i] for i in range(n_z)] #volumetric flows (m3/s)
    plenum_flow = [[0 for j in range(n_z)] for i in range(n_z)] #keep track of what zone plenum flow is from
    inflow = [0 for i in range(n_p)]
    for i in range(n_p):
        z=building.hvac['plenum']['zone_index'][i]
        for j in building.hvac['plenum']['in_zones'][i]:
            inflow[i] += return_flow[j]
            plenum_flow[z][j] =return_flow[j]
            return_flow[j] = 0
        return_flow[z] +=inflow[i]
    flow_exiting_zone = [hvac_supply[i] + sum(mixing[i]) + infiltration[i] + sum(plenum_flow[i])  for i in range(n_z)]#flow leaves through exhaust, mixing to another zone or plenum, or return air
    return flow_exiting_zone,return_flow,plenum_flow

def return_air_calc(building,gains,T,w,return_flow,direct_flow,direct_flow_fresh):
    #determine air returning to HVAC from zone, and agregate into loops
    air_density = 1.204 #kg/m^3  #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    n_z = len(T)
    #add sensible and latent heat from refrigeration equipment
    latent_H2O = [gains['hvac_latent'][i]/2264705  for i in range(n_z)]#convert J/s to kg/s of water
    air_H2O = [w[i]*return_flow[i]*air_density for i in range(n_z)]
    cp_zone = [1006 + 1860*w[i] for i in range(n_z)] #J/kg*K
    return_air = {}
    return_air['m_dot'] = [return_flow[i]*air_density + latent_H2O[i] for i in range(n_z)] #mass flow (kg/s)
    return_air['Tdb'] = [0 for i in range(n_z)]
    return_air['w'] = [0 for i in range(n_z)]
    for i in range(n_z):
        if return_air['m_dot'][i]>0:
            return_air['Tdb'][i] = T[i] + gains['hvac_sensible'][i]/(cp_zone[i]*return_air['m_dot'][i])
            return_air['w'][i] = (air_H2O[i] + latent_H2O[i])/return_air['m_dot'][i]
        else:
            return_air['Tdb'][i] = T[i] 
            return_air['w'][i] = w[i] 
    return_air['h'] =  psychometric(return_air,'h')

    central_return_air = {}
    central_return_air['Tdb'] = [j for j in return_air['Tdb']]
    central_return_air['w'] = [j for j in return_air['w']]
    central_return_air['h'] = [j for j in return_air['h']]
    central_return_air['m_dot'] = [max(0,return_air['m_dot'][i] - (direct_flow[i] - direct_flow_fresh[i])*air_density) for i in range(n_z)]

    direct_return_air = {}
    direct_return_air['Tdb'] = [j for j in return_air['Tdb']]
    direct_return_air['w'] = [j for j in return_air['w']]
    direct_return_air['h'] = [j for j in return_air['h']]
    direct_return_air['m_dot'] = [max(0,(direct_flow[i] - direct_flow_fresh[i])*air_density) for i in range(n_z)]

    #flow agregated into air loops for supply side
    loop_return_air = {}
    n_l = len(building.hvac['loop']['name'])
    loop_return_air['m_dot'] = [0 for i in range(n_l)]
    loop_return_air['h'] = [0 for i in range(n_l)]
    loop_return_air['w'] = [0 for i in range(n_l)]
    for i in range(n_z):
        if building.zones['air_loop'][i]!=None:
            loop_return_air['m_dot'][building.zones['air_loop'][i]] += central_return_air['m_dot'][i]*building.zones['multiplier'][i]    
            loop_return_air['h'][building.zones['air_loop'][i]] += central_return_air['m_dot'][i]*central_return_air['h'][i]*building.zones['multiplier'][i]
            loop_return_air['w'][building.zones['air_loop'][i]] += central_return_air['m_dot'][i]*central_return_air['w'][i]*building.zones['multiplier'][i]
    for i in range(n_l):
        if loop_return_air['m_dot'][i]>0:
            loop_return_air['h'][i] = loop_return_air['h'][i]/loop_return_air['m_dot'][i]
            loop_return_air['w'][i] = loop_return_air['w'][i]/loop_return_air['m_dot'][i]
        else: #avoid issues with zero or nearly zero flow
            onl = len([j for j in range(n_z) if building.zones['air_loop'][j]==i])
            if onl>0:
                loop_return_air['h'][i] = sum([central_return_air['h'][j] for j in range(n_z) if building.zones['air_loop'][j]==i])/onl
                loop_return_air['w'][i] = sum([central_return_air['w'][j] for j in range(n_z) if building.zones['air_loop'][j]==i])/onl
    loop_return_air['Tdb'] =  psychometric(loop_return_air,'Tdb')
    return direct_return_air,loop_return_air

def mix_air(s1,s2):
    mixed_air = {}
    mixed_air['m_dot'] = s1['m_dot'] + s2['m_dot'] #mass flow (kg/s)
    if mixed_air['m_dot']==0:
        for k in s1:
            mixed_air[k] = s1[k]+0
    else:
        mixed_air['h'] = (s1['m_dot']*s1['h'] + s2['m_dot']*s2['h'])/mixed_air['m_dot']
        w_sat = psychometric({'h':mixed_air['h']},'w')
        mix_w = (s1['m_dot']*s1['w'] + s2['m_dot']*s2['w'])/mixed_air['m_dot']
        mixed_air['m_dot'] = mixed_air['m_dot']*(1 - max(0,mix_w-w_sat))
        mixed_air['w'] = min(w_sat,mix_w)
        mixed_air['Tdb'] = psychometric(mixed_air,'Tdb')
    return mixed_air

def loop_supply_setpoint(building,profile,air_nodes,fresh_air,fresh_air_loop,loop_return_air,loop_flow,cf,loop_2_zone,cp_zone,flow_exiting_zone,net_H2O_gain,T_set,w_set,Q_Tset,T_min,T_max,fan_avail,des_day):
    air_density = 1.204 #kg/m^3  #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    n_z = len(cf)
    loop_T_supply = [j for j in loop_return_air['Tdb']]
    loop_w_supply = [j for j in loop_return_air['w']]
    mixed_air_loop = {}
    for k in loop_return_air:
        mixed_air_loop[k] = [j for j in loop_return_air[k]]
    for i in range(len(fresh_air_loop)):
        term_i = [j for j ,x in enumerate(building.hvac['terminals']['loop']) if x==i]
        z_i = [building.hvac['terminals']['zone'][j]  for j in term_i]
        # cf = [central_flow[j]  for j in z_i]
        on_loop = [j for j ,x in enumerate(building.air_supply_equip['loop']) if x == i]
        if sum([cf[j]  for j in z_i]) > 0:
            air={}
            air['Tdb'] = loop_return_air['Tdb'][i]+0
            air['w'] = loop_return_air['w'][i]+0
            air['h'] = loop_return_air['h'][i]+0
            air['m_dot'] = loop_return_air['m_dot'][i]+0
            for j in on_loop:
                if building.air_supply_equip['type'][j][:4]=='fan:':
                    fan_name = building.air_supply_equip['name'][j]
                elif building.air_supply_equip['type'][j]=='airloophvac:unitaryheatcool':
                    fan_name = building.unitary_heat_cool[building.air_supply_equip['name'][j]]['fan_name']
            sched = False
            try:
                k = building.manager['node'].index(building.air_supply_nodes['name'][building.air_supply_equip['outlet'][on_loop[-1]]])
                if building.manager['type'][k]=='Scheduled':
                    sched = True
                    for j in on_loop:
                        out = building.air_supply_equip['outlet'][j]
                        node = building.air_supply_nodes['name'][out]
                        if building.air_supply_equip['type'][j]=='airloophvac:outdoorairsystem':
                            mixed_air,fresh_air_loop[i] = economized_outdoor_air(building,fresh_air_loop[i],air,fresh_air,loop_flow[i],i,fan_avail,fan_name,des_day)
                            mixed_air_loop['Tdb'][i] = air['Tdb']+0
                            mixed_air_loop['w'][i] = air['w']+0
                            mixed_air_loop['h'][i] = air['h']+0
                            mixed_air_loop['m_dot'][i] = air['m_dot']+0
                            air_nodes['supply_T_set'][out] = mixed_air['Tdb']+0
                        else:
                            air_nodes['supply_T_set'][out] = manager_override(building,profile,node,profile[building.manager['schedule'][k]],air,fresh_air,fan_avail)
            except ValueError:
                pass
            if not sched:
                _,_,Q_to_air = fan_calc(building.fans[fan_name],air,fan_avail[fan_name])
                Q_f2zone = [0 for x in range(n_z)]
                for x in z_i:
                    Q_f2zone[x] = Q_to_air*loop_2_zone[x][i]/building.zones['multiplier'][x] #fan heat to zone air supply
                air_nodes['supply_T_set'][building.air_supply_equip['inlet'][on_loop[0]]] = air['Tdb']
                for j  in range(len(on_loop)):
                    ## making sure simulation of these components is in the order they are on the supply branch
                    for k in on_loop:
                        if building.air_supply_equip['branch_order'][k] == j:
                            out = building.air_supply_equip['outlet'][k]
                            inl = building.air_supply_equip['inlet'][k]
                            ty = building.air_supply_equip['type'][k]
                            break
                    
                    air_nodes['supply_T_set'][out] = air_nodes['supply_T_set'][inl]
                    node = building.air_supply_nodes['name'][out]
                    if ty=='airloophvac:outdoorairsystem':
                        air,fresh_air_loop[i] = economized_outdoor_air(building,fresh_air_loop[i],air,fresh_air,loop_flow[i],i,fan_avail,fan_name,des_day)
                        air_nodes['supply_T_set'][out] = air['Tdb']+0
                        mixed_air_loop['Tdb'][i] = air['Tdb']+0
                        mixed_air_loop['w'][i] = air['w']+0
                        mixed_air_loop['h'][i] = air['h']+0
                        mixed_air_loop['m_dot'][i] = air['m_dot']+0
                    elif ty[:6]=='coil:h':
                        if any([(cf[x]*cp_zone[x]*(air['Tdb']-T_set['heat'][x])<(Q_Tset[1][x]-Q_f2zone[x]) and cf[x]>0) for x in z_i]): #HVAC in heating mode
                            air['Tdb'] = -1e4 #set arbitraily small so that it gets set to maximum of z_sup_T (may be smaller than initial air['Tdb'])
                            for x in z_i:
                                if cf[x]>0:
                                    z_sup_T = Q_Tset[1][x]/(cp_zone[x]*cf[x]) + T_set['heat'][x]
                                    air['Tdb'] = max(air['Tdb'],min(max(z_sup_T,T_min[x]),T_max[x]))
                            air['Tdb'] = min(air['Tdb'],building.hvac['design']['Tsupply_h'][i]) #constrain supply temperature
                            air_nodes['supply_T_set'][out] = manager_override(building,profile,node,air['Tdb'],air,fresh_air,fan_avail)
                    elif ty[:4]=='coil' or ty[:3]=='dx:':
                        if any([(cf[x]*cp_zone[x]*(air['Tdb']-T_set['cool'][x])>(Q_Tset[0][x]-Q_f2zone[x]) and cf[x]>0) for x in z_i]): #HVAC in cooling mode
                            air['Tdb'] = 1e4 #set arbitraily large so that it gets set to minimum of z_sup_T (may be larger than initial air['Tdb'])
                            for x in z_i:
                                if cf[x]>0:
                                    z_sup_T = Q_Tset[0][x]/(cp_zone[x]*cf[x]) + T_set['cool'][x]
                                    air['Tdb'] = max(min(air['Tdb'],z_sup_T),building.hvac['design']['Tsupply_c'][i]) #constrain supply temperature
                            air_nodes['supply_T_set'][out] = manager_override(building,profile,node,air['Tdb'],air,fresh_air,fan_avail)
                        if any([w_set['dehumidify'][x]<1 for x in z_i]):
                            w_target = psychometric({'Tdb':[T_set['cool'][x] for x in z_i],'RH':[w_set['dehumidify'][x]*100 for x in z_i]},'w')
                            w_supply = max(.001,sum([flow_exiting_zone[x]*w_target - net_H2O_gain[x] for x in z_i])/sum([cf[x]  for x in z_i])) # find supply humidity
                            air['w'] = w_supply+0
                            if mixed_air_loop['w'][i] > air['w']:
                                air['Tdb'] =  min(psychometric({'Tdb':[T_set['cool'][x] for x in z_i],'w':[w_supply]},'Tdp'))
                                air_nodes['supply_T_set'][out] = manager_override(building,profile,node,air['Tdb'],air,fresh_air,fan_avail)
                    elif ty[:4]=='fan:':
                        _,_,Q_to_air = fan_calc(building.fans[fan_name],air,fan_avail[fan_name])
                        air['Tdb'] = air_nodes['supply_T_set'][inl] + Q_to_air/(air['m_dot']*sum([cp_zone[x] for x in z_i])/len(z_i)/air_density)
                        air_nodes['supply_T_set'][out] = air['Tdb']+0 #manager_override(building,profile,node,air['Tdb'],air,fresh_air,fan_avail)
                        Q_f2zone = [0 for i in range(n_z)] #zero out for any components downstream of fan
                    elif ty=='airloophvac:unitaryheatcool':
                        if any([(cf[x]*cp_zone[x]*(air['Tdb']-T_set['cool'][x])>(Q_Tset[0][x]-Q_f2zone[x]) and cf[x]>0) for x in z_i]): #HVAC in cooling mode
                            air['Tdb'] = 1e4 #set arbitraily large so that it gets set to minimum of z_sup_T (may be larger than initial air['Tdb'])
                            for x in z_i:
                                if cf[x]>0:
                                    z_sup_T = Q_Tset[0][x]/(cp_zone[x]*cf[x]) + T_set['cool'][x]
                                    air['Tdb'] = min(air['Tdb'],z_sup_T)
                        elif any([(cf[x]*cp_zone[x]*(air['Tdb']-T_set['heat'][x])<(Q_Tset[1][x]-Q_f2zone[x]) and cf[x]>0) for x in z_i]): #HVAC in heating mode
                            air['Tdb'] = -1e4 #set arbitraily small so that it gets set to maximum of z_sup_T (may be smaller than initial air['Tdb'])
                            for x in z_i:
                                if cf[x]>0:
                                    z_sup_T = Q_Tset[1][x]/(cp_zone[x]*cf[x]) + T_set['heat'][x]
                                    air['Tdb'] = max(air['Tdb'],min(max(z_sup_T,T_min[x]),T_max[x]))
                            air['Tdb'] = min(building.hvac.design.Tsupply_h[i],air['Tdb'])
                        else:
                            air = fan_calc(building.fans[fan_name],air,fan_avail[fan_name])
                        air_nodes['supply_T_set'][out] = manager_override(building,profile,building.air_supply_nodes.name[inl],air['Tdb'],air,fresh_air,fan_avail)
                    elif ty=='humidifier:steam:electric':
                        if any([w_set['humidify'][x]>0 for x in z_i]):
                            ##need to figure out temperature when humidifying?
                            w_target = psychometric({'Tdb':[T_set['heat'][x] for x in z_i],'RH':[w_set['humidify'][x]*100 for x in z_i]},'w') #heating = maybe humidifying
                            w_supply = max(.001,sum([flow_exiting_zone[z_i[x]]*w_target[x] - net_H2O_gain[z_i[x]] for x in range(len(z_i))])/sum([cf[x]  for x in z_i])) # find supply humidity
                            if mixed_air_loop['w'][i]<w_supply:
                                air['w'] = max(w_supply)
                        air_nodes['supply_T_set'][out] = air['Tdb']+0
                    else:
                        print('something else on air loop?')
            loop_T_supply[i] = air_nodes['supply_T_set'][out]+0
            loop_w_supply[i] = air['w']+0
        else:
            for j in on_loop:
                out = building.air_supply_equip['outlet'][j]
                air_nodes['supply_T_set'][out] = loop_return_air['Tdb'][i]+0
    return loop_T_supply,loop_w_supply,mixed_air_loop,air_nodes

def economized_outdoor_air(building,min_flow,air,fresh_air,loop_flow,l,fan_avail,fan_name,des_day):
    air_density = 1.204 #kg/m^3 %standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    r_air = {}
    r_air['Tdb'] = air['Tdb']+0
    r_air['w'] = air['w']+0
    r_air['h'] = air['h']+0
    f_air = {}
    f_air['Tdb'] = fresh_air['Tdb']+0
    f_air['w'] = fresh_air['w']+0
    f_air['h'] = fresh_air['h']+0
    fresh_air_loop = min_flow+0
    T_supply_max = building.hvac['design']['Tsupply_h'][l]
    Ta = f_air['Tdb']+0
    Tb = r_air['Tdb']+0
    ## run checks to see if economizer is active
    econ_on = False
    if econ_on and not des_day and T_supply_max>Ta and r_air['Tdb']>T_supply_max:
        error = 1
        max_flow = building.hvac.outdoor_air.max_flow(l)
        while error>1e-3:
            f_air['m_dot'] = fresh_air_loop*air_density
            r_air['m_dot'] = (loop_flow - fresh_air_loop)*air_density
            mixed_air = mix_air(r_air,f_air)
            cp = (1006 + 1860*mixed_air['w']) #Specific heat in kJ/kg
            _,_,Q_to_air = fan_calc(building.fans[fan_name],mixed_air,fan_avail[fan_name])
            T_no_treat = mixed_air['Tdb'] + Q_to_air/(mixed_air['m_dot']*cp)
            error = T_no_treat - T_supply_max
            if (error>0 and fresh_air_loop>=max_flow) or (error<0 and fresh_air_loop<=min_flow):
                error = 0
            d_flow = 0
            if (Ta-Tb)!=0:
                d_flow = loop_flow*(1-(Tb+20)/(error+20))/(Ta-Tb)*error
            fresh_air_loop = max(min_flow,min(max_flow,fresh_air_loop + d_flow))
    else:
        f_air['m_dot'] = fresh_air_loop*air_density
        r_air['m_dot'] = (loop_flow - fresh_air_loop)*air_density
        mixed_air = mix_air(r_air,f_air)
    return mixed_air,fresh_air_loop

def zones_supply_setpoint(building,T,w,central_flow,cp_zone,loop_T_supply,direct_return_air,T_set,Q_Tset,T_min,T_max,fresh_air,direct_fresh_air,direct_flow,fan_avail):
    air_density = 1.204 #kg/m^3 #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    n_z = len(building.zones['name'])
    supply_flow = {}
    supply_flow['Tdb'] = [T['zone'][i] for i in range(n_z)]
    for z in range(n_z):
        if central_flow[z]>0:
            try:
                term_z = building.hvac['terminals']['zone'].index(z)
                supply_flow['Tdb'][z] = loop_T_supply[building.hvac['terminals']['loop'][term_z]]
                if building.hvac['terminals']['type'][term_z] == 'SingleDuct:Uncontrolled':
                    pass
                    ##do nothing
                elif building.hvac['terminals']['type'][term_z] == 'SingleDuct:ConstantVolume:Reheat':
                    ##need example
                    pass
                elif building.hvac['terminals']['type'][term_z] == 'SingleDuct:VAV:Reheat':
                    if central_flow[z]*cp_zone[z]*(supply_flow['Tdb'][z]- T_set['heat'][z])<Q_Tset[1][z]: #needs heat
                        m = building.zones['multiplier'][z]
                        Q_reheat_max = m*central_flow[z]*cp_zone[z]*(T_max[z] - supply_flow['Tdb'][z])
                        # Q_reheat_max = min(Q_reheat_max,building.coils_heating[building.hvac['terminals']['reheat_coil_name'][term_z]]['capacity']) #remove, sometimes coil exceeds rated maximum
                        Q_reheat_for_target = m*(Q_Tset[1][z] - central_flow[z]*cp_zone[z]*(supply_flow['Tdb'][z]- T_set['heat'][z]))
                        Q_terminal = max(0,min(Q_reheat_for_target,Q_reheat_max))
                        supply_flow['Tdb'][z] += Q_terminal/(central_flow[z]*m*cp_zone[z])
                else:
                    print('air terminal of type: '+ building.hvac['terminals']['type'][term_z])
            except ValueError:
                #not on an air loop (no terminal) check unitary systems
                supply_flow['Tdb'][z]= direct_return_air['Tdb'][z]
    supply_flow['w'] = [j for j in w['zone']]
    for i in range(n_z):
        if building.zones['air_loop'][i]!=None:
            supply_flow['w'][i] =w['loop'][building.zones['air_loop'][i]]
    
    supply_flow['flow'] = [j for j in central_flow]
    supply_flow['m_dot'] = [central_flow[x]*air_density for x in range(len(central_flow))]
    supply_flow['h'] = psychometric(supply_flow,'h')
    T['central'] = [j for j in supply_flow['Tdb']]
    T['direct'] = [j for j in T['central']]
    w['central'] = [j for j in supply_flow['w']]
    w['direct'] = [j for j in w['central']]
    
    for i in building.unitary_sys:
        z = building.unitary_sys[i]['zone']
        fan_name = building.unitary_sys[i]['fan_name']
        if direct_flow[z]>0:
            c_air = {}
            for x in supply_flow:
                c_air[x] = supply_flow[x][z]
            r_air = {}
            for x in direct_return_air:
                r_air[x] = direct_return_air[x][z]
            if 'terminal' in building.unitary_sys[i]:
                r_air['m_dot'] = direct_flow[z]*air_density - c_air['m_dot']
                _,_,Q_to_air = fan_calc(building.fans[fan_name],mixed_air,fan_avail[fan_name]) #unitary system fan is for all zones (divide by multiplier)
                mixed_air = mix_air(r_air,c_air)
                Q_direct = [Q_Tset[0][z],Q_Tset[1][z]]
            else:
                r_air['m_dot'] = (direct_flow[z] - direct_fresh_air[z])*air_density 
                Q_loop_air = supply_flow['flow'][z]*cp_zone[z]*(supply_flow['Tdb'][z] - T['zone_est'][z])
                f_air = {}
                for k in fresh_air:
                    f_air[k] = fresh_air[k]+0
                f_air['m_dot'] = direct_fresh_air[z]*air_density
                mixed_air = mix_air(r_air,f_air) #fresh air directly to zone, not outdoor air mixing on loop
                Q_direct = [Q_Tset[0][z]+Q_loop_air,Q_Tset[1][z]+Q_loop_air]
                _,_,Q_to_air = fan_calc(building.fans[fan_name],mixed_air,fan_avail[fan_name]) #unitary system fan is for all zones (divide by multiplier)


            supply_flow['Tdb'][z] = mixed_air['Tdb'] + Q_to_air/(mixed_air['m_dot']*cp_zone[z]*building.zones['multiplier'][z])
            supply_flow['w'][z] = mixed_air['w']
            supply_flow['m_dot'][z] = mixed_air['m_dot']
            supply_flow['h'][z] = mixed_air['h']
            flow = mixed_air['m_dot']/air_density

            if building.unitary_sys[i]['type']=='PackagedTerminalAirConditioner' or building.unitary_sys[i]['type']=='FourPipeFanCoil':
                if flow*cp_zone[z]*(supply_flow['Tdb'][z] - T_set['cool'][z])>Q_direct[0]: #unitary system in cooling mode
                    T_unit_cooler = min(max(Q_direct[0]/(cp_zone[z]*flow) + T_set['cool'][z],T_min[z]),T_max[z])
                    supply_flow['Tdb'][z] = min(supply_flow['Tdb'][z],T_unit_cooler)
                elif flow*cp_zone[z]*(supply_flow['Tdb'][z] - T_set['heat'][z])<Q_direct[1]: #unitary system  in heating mode
                    T_unit_heater = min(Q_direct[1]/(cp_zone[z]*flow) + T_set['heat'][z],T_max[z])
                    supply_flow['Tdb'][z] = max(supply_flow['Tdb'][z],T_unit_heater)
            elif building.unitary_sys[i]['type']=='UnitHeater':
                if flow*cp_zone[z]*(supply_flow['Tdb'][z] - T_set['heat'][z])<Q_direct[1]: #unitary system  in heating mode
                    T_unit_heater = min(Q_direct[1]*(cp_zone[z]*flow) + T_set['heat'][z],T_max[z])
                    supply_flow['Tdb'][z] = max(supply_flow['Tdb'][z],T_unit_heater)
            elif building.unitary_sys[i]['type']=='WindowAirConditioner':
                print('need WindowAirConditioner')
            else:
                print('need other types of unitary system')
            T['direct'][z] = supply_flow['Tdb'][z]
            w['direct'][z] = supply_flow['w'][z]
            if 'terminal' not in building.unitary_sys[i]: #Mixing central air for net supply to zone
                d_air = {}
                d_air['Tdb'] = supply_flow['Tdb'][z]
                d_air['w'] = supply_flow['w'][z]
                d_air['m_dot'] = supply_flow['m_dot'][z]
                d_air['h'] = supply_flow['h'][z]
                mixed_air = mix_air(d_air,c_air)
                supply_flow['Tdb'][z] = mixed_air['Tdb'] 
                supply_flow['w'][z] = mixed_air['w']
                supply_flow['m_dot'][z] = mixed_air['m_dot']
                supply_flow['h'][z] = mixed_air['h']
            supply_flow['flow'] = [supply_flow['m_dot'][i]/air_density for i in range(n_z)]
    return supply_flow, T, w

def des_day_zone_flow(building,fresh_air_zone,cp_supply,Q_Tset,T_set,T_min,T_max,flow_imbalance):
    n_z = len(building.zones['name'])
    zone_flow_req = [0 for i in range(len(n_z))]
    for i in range(n_z):
        if T_set['no_hvac'][i]:
            pass
        elif Q_Tset[1][i]>0:
            zone_flow_req[i] = max(Q_Tset[1][i]/(cp_supply[i]*(T_max[i] - T_set['heat'][i])),-flow_imbalance[i])
        elif Q_Tset[0][i]<0:
            zone_flow_req[i] =  max(Q_Tset[0][i]/(cp_supply[i]*(T_min[i] - T_set['cool'][i])),-flow_imbalance[i])
    zone_flow = [max(fresh_air_zone[i],zone_flow_req[i]) for i in range(n_z)]#volumetric flow m^3/s
    direct_flow = [0 for i in range(n_z)] #flow for zones not on a HVAC loop (unit heaters)
    for i in range(len(building.unitary_sys)):
        try:
            l = building.hvac['loop']['supply_outlet'].index(building.unitary_sys[i]['outlet'])
            z = building.unitary_sys[i]['zone']
            direct_flow[z] = zone_flow[z]
        except ValueError:
            pass
    central_flow = [zone_flow[i] - direct_flow[i] for i in range(n_z)]
    n_l = len(building.hvac['loop']['name'])
    supply_flow = [0 for i in range(n_l)]
    fresh_air_loop= [0 for i in range(n_l)]
    for i in range(n_z):
        supply_flow[building.zones['air_loop'][i]] += central_flow[i]*building.zones['multiplier'][i] #mass flow for the loop (kg/s)
        fresh_air_loop[building.zones['air_loop'][i]] += fresh_air_zone[i]
    loop_2_zone = []
    zone_request= [[0 for j in range(n_l)] for i in range(n_z)]
    loop_request = [0 for i in range(n_l)]
    for i in range(n_z):
        l=building.zones['air_loop'][i]
        if l!=None:
            zone_request[i][l]= central_flow[i]*building.zones['multiplier'][i]
            loop_request[l] += zone_request[i][l]
    for i in range(n_z):
        if loop_request[l]>0:
            loop_2_zone.append([zone_request[i][l]/loop_request[l] for l in range(n_l)])
        else:
            loop_2_zone.append([0  for l in range(n_l)])
    direct_fresh_air = [fresh_air_zone[i] - sum([fresh_air_loop[l]*loop_2_zone[i][l] for l in range(n_l)]) for i in range(n_z)]
    return central_flow,direct_flow,supply_flow,direct_fresh_air,fresh_air_loop


"""
lowess: Locally linear regression
==================================
Implementation of the LOWESS algorithm in n dimensions.
References
=========
[HTF] Hastie, Tibshirani and Friedman (2008). The Elements of Statistical
Learning; Chapter 6
[Cleveland79] Cleveland (1979). Robust Locally Weighted Regression and Smoothing
Scatterplots. J American Statistical Association, 74: 829-836.
"""
import numpy as np
import numpy.linalg as la


# Kernel functions:
def epanechnikov(xx, **kwargs):
    """
    The Epanechnikov kernel estimated for xx values at indices idx (zero
    elsewhere)
    Parameters
    ----------
    xx: float array
        Values of the function on which the kernel is computed. Typically,
        these are Euclidean distances from some point x0 (see do_kernel)
    Notes
    -----
    This is equation 6.4 in HTF chapter 6
    """
    l = kwargs.get('l', 1.0)
    ans = np.zeros(xx.shape)
    xx_norm = xx / l
    idx = np.where(xx_norm <= 1)
    ans[idx] = 0.75 * (1 - xx_norm[idx]  ** 2)
    return ans


def tri_cube(xx, **kwargs):
    """
    The tri-cube kernel estimated for xx values at indices idx (zero
    elsewhere)
    Parameters
    ----------
    xx: float array
        Values of the function on which the kernel is computed. Typically,
        these are Euclidean distances from some point x0 (see do_kernel)
    idx: tuple
        An indexing tuple pointing to the coordinates in xx for which the
        kernel value is estimated. Default: None (all points are used!)
    Notes
    -----
    This is equation 6.6 in HTF chapter 6
    """
    ans = np.zeros(xx.shape)
    idx = np.where(xx <= 1)
    ans[idx] = (1 - np.abs(xx[idx]) ** 3) ** 3
    return ans


def bi_square(xx, **kwargs):
    """
    The bi-square weight function calculated over values of xx
    Parameters
    ----------
    xx: float array
    Notes
    -----
    This is the first equation on page 831 of [Cleveland79].
    """
    ans = np.zeros(xx.shape)
    idx = np.where(xx < 1)
    ans[idx] = (1 - xx[idx] ** 2) ** 2
    return ans


def do_kernel(x0, x, l=1.0, kernel=epanechnikov):
    """
    Calculate a kernel function on x in the neighborhood of x0
    Parameters
    ----------
    x: float array
       All values of x
    x0: float
       The value of x around which we evaluate the kernel
    l: float or float array (with shape = x.shape)
       Width parameter (metric window size)
    kernel: callable
        A kernel function. Any function with signature: `func(xx)`
    """
    # xx is the norm of x-x0. Note that we broadcast on the second axis for the
    # nd case and then sum on the first to get the norm in each value of x:
    xx = np.sqrt(np.sum(np.power(x - x0[:, np.newaxis], 2), 0))
    return kernel(xx, l=l)


def lowess(x, y, x0, deg=1, kernel=epanechnikov, l=1, robust=False,):
    """
    Locally smoothed regression with the LOWESS algorithm.
    Parameters
    ----------
    x: float n-d array
       Values of x for which f(x) is known (e.g. measured). The shape of this
       is (n, j), where n is the number the dimensions and j is the
       number of distinct coordinates sampled.
    y: float array
       The known values of f(x) at these points. This has shape (j,)
    x0: float or float array.
        Values of x for which we estimate the value of f(x). This is either a
        single scalar value (only possible for the 1d case, in which case f(x0)
        is estimated for just that one value of x), or an array of shape (n, k).
    deg: int
        The degree of smoothing functions. 0 is locally constant, 1 is locally
        linear, etc. Default: 1.
    kernel: callable
        A kernel function. {'epanechnikov', 'tri_cube', 'bi_square'}
    l: float or float array with shape = x.shape
       The metric window size for the kernel
    robust: bool
        Whether to apply the robustification procedure from [Cleveland79], page
        831
    Returns
    -------
    The function estimated at x0.
    Notes
    -----
    The solution to this problem is given by equation 6.8 in Hastie
    Tibshirani and Friedman (2008). The Elements of Statistical Learning
    (Chapter 6).
    Example
    -------
    >>> import lowess as lo
    >>> import numpy as np
    # For the 1D case:
    >>> x = np.random.randn(100)
    >>> f = np.cos(x) + 0.2 * np.random.randn(100)
    >>> x0 = np.linspace(-1, 1, 10)
    >>> f_hat = lo.lowess(x, f, x0)
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots(1)
    >>> _ = ax.scatter(x, f)
    >>> _ = ax.plot(x0, f_hat, 'ro')
    # 2D case (and more...)
    >>> x = np.random.randn(2, 100)
    >>> f = -1 * np.sin(x[0]) + 0.5 * np.cos(x[1]) + 0.2*np.random.randn(100)
    >>> x0 = np.mgrid[-1:1:.1, -1:1:.1]
    >>> x0 = np.vstack([x0[0].ravel(), x0[1].ravel()])
    >>> f_hat = lo.lowess(x, f, x0, kernel=lo.tri_cube)
    >>> from mpl_toolkits.mplot3d import Axes3D
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection='3d')
    >>> _ = ax.scatter(x[0], x[1], f)
    >>> _ = ax.scatter(x0[0], x0[1], f_hat, color='r')
    """
    if robust:
        # We use the procedure described in Cleveland1979
        # Start by calling this function with robust set to false and the x0
        # input being equal to the x input:
        y_est = lowess(x, y, x, kernel=epanechnikov, l=l, robust=False)
        resid = y_est - y
        median_resid = np.nanmedian(np.abs(resid))
        # Calculate the bi-cube function on the residuals for robustness
        # weights:
        robustness_weights = bi_square(resid / (6 * median_resid))

    # For the case where x0 is provided as a scalar:
    if not np.iterable(x0):
       x0 = np.asarray([x0])
    ans = np.zeros(x0.shape[-1])
    # We only need one design matrix for fitting:
    B = [np.ones(x.shape[-1])]
    for d in range(1, deg+1):
        B.append(x ** deg)

    B = np.vstack(B).T
    for idx, this_x0 in enumerate(x0.T):
        # This is necessary in the 1d case (?):
        if not np.iterable(this_x0):
            this_x0 = np.asarray([this_x0])
        # Different weighting kernel for each x0:
        W = np.diag(do_kernel(this_x0, x, l=l, kernel=kernel))
        # XXX It should be possible to calculate W outside the loop, if x0 and
        # x are both sampled in some regular fashion (that is, if W is the same
        # matrix in each iteration). That should save time.

        if robust:
            # We apply the robustness weights to the weighted least-squares
            # procedure:
            robustness_weights[np.isnan(robustness_weights)] = 0
            W = np.dot(W, np.diag(robustness_weights))
        #try:
        # Equation 6.8 in HTF:
        BtWB = np.dot(np.dot(B.T, W), B)
        BtW = np.dot(B.T, W)
        # Get the params:
        beta = np.dot(np.dot(la.pinv(BtWB), BtW), y.T)
        # We create a design matrix for this coordinat for back-predicting:
        B0 = [1]
        for d in range(1, deg+1):
            B0 = np.hstack([B0, this_x0 ** deg])
        B0 = np.vstack(B0).T
        # Estimate the answer based on the parameters:
        ans[idx] += np.dot(B0, beta)
    # If we are trying to sample far away from where the function is
    # defined, we will be trying to invert a singular matrix. In that case,
    # the regression should not work for you and you should get a nan:
        #except la.LinAlgError :
        #    ans[idx] += np.nan
    return ans.T


from eagers.config.network import NETWORK_NAMES, NETWORK_ABBRS


def case_marginal(dispatch, gen, qp, v_h):
    # Find all ouptuts to have marginal cost of.
    out_names = list(qp['organize']['balance'].keys())
    outs = []
    marginal = {}
    for k in range(len(NETWORK_NAMES)):
        if NETWORK_NAMES[k] in out_names:
            outs.append(NETWORK_ABBRS[k])
            marginal[NETWORK_ABBRS[k]] = None

    # If storage, use marginal cost from dispatch of storage.
    ac_dc = None
    n_g = len(gen)
    for i in range(n_g):
        if gen[i]['type'] == 'ACDCConverter':
            ac_dc = gen[i]
    for i in range(n_g):
        if gen[i]['type'] == 'ElectricStorage':
            s = qp['indices'][i][0][0]
            mar = dispatch[i] * qp['h'][s] + qp['f'][s]
            if 'e' in outs:
                if not ac_dc is None and 'dc' in gen[i]['output']:
                    marginal['e'] = mar / ac_dc['output']['e'][0][1]
                else:
                    marginal['e'] = mar
            if 'dc' in outs:
                if not ac_dc is None and 'e' in gen[i]['output']:
                    marginal['dc'] = mar / ac_dc['output']['dc'][0][0]
                else:
                    marginal['dc'] = mar
        if gen[i]['type'] == 'ThermalStorage':
            s = qp['indices'][i][0][0]
            mar = dispatch[i] * qp['h'][s] + qp['f'][s]
            f = list(gen[i]['output'].keys())
            marginal[f[0]] = mar
    # Otherwise find marginal cost of each active generator with this
    # output and take max.
    for k, out in enumerate(outs):
        if marginal[out] is None:
            for i in range(n_g):
                d = dispatch[i]
                s = qp['indices'][i][0]
                if d > 0 and 'output' in gen[i] and not s is None:
                    f = list(gen[i]['output'].keys())
                    if out in f and all([output>0 for output in gen[i]['output'][out][-1]]):
                        j = 0
                        while d > 0 and j < len(s):
                            seg = min([qp['ub'][s[k]], d])
                            mar = dispatch[i] * qp['h'][s[k]] + qp['f'][s[k]]
                            d -= seg
                            j += 1
                        if mar == 0:
                            s2 = ['e','dc','h','cw','hy']
                            for k2 in s2:
                                if not marginal[k2] is None and k2 in gen[i]['output']:
                                    mar = max(-output * marginal[k2] for output in gen[i]['output'][k2][-1])
                                    break
                        if gen[i]['type'] == 'CombinedHeatPower':
                            if out == 'h':
                                mar = mar * 0.3
                            elif v_h and (out == 'e' or out == 'dc'):
                                mar = mar * 0.7
                        if marginal[out] is None:
                            marginal[out] = mar
                        else:
                            marginal[out] = max([marginal[outs[k]], mar])
        if marginal[out] is None:
            marginal[out] = 1  # Should only reach this if there is no storage an demand = 0 (only way for zero generators to be active).
    return marginal



def best_eff(gen, scale_cost, marginal):
    '''
    identifies the operating point of maximum efficiency (lowest cost per kW)
    used to estimate the lowest possible cost if a generator is added 
    in order to eliminate expensive cases without testing them
    '''
    n_g = len(gen)
    c_red = {}
    c_red['eff'] = [0 for i in range(n_g)]
    c_red['p'] = [0 for i in range(n_g)]
    c_red['c'] = [0 for i in range(n_g)]
    c_red['ab'] = [None for i in range(n_g)]
    for i in range(n_g):
        if gen[i]['type'] in ['CombinedHeatPower', 'ElectricGenerator', 'HydrogenGenerator','Heater','Chiller']:
            c_red['eff'][i] = gen[i]['max_eff']
            c_red['p'][i] = gen[i]['max_eff_point']
            c_red['c'][i] = scale_cost[i][0]
        if gen[i]['type'] in ['CombinedHeatPower', 'ElectricGenerator']:
            if 'e' in gen[i]['output']:
                c_red['ab'][i] = 'e'
            elif 'dc' in gen[i]['output']: 
                c_red['ab'][i] = 'dc'
        if gen[i]['type'] == 'Chiller':
            c_red['ab'][i] = 'c'
            if gen[i]['source'] == 'heat':
                c_red['c'][i] = marginal['h']
            else:
                c_red['c'][i] = marginal['e']
        if gen[i]['type'] == 'Heater':
            c_red['ab'][i] = 'h'
        if gen[i]['type'] in  ['HydrogenGenerator','CoolingTower','Electrolyzer']:
            c_red['c'][i] = marginal['e']
            c_red['ab'][i] = 'hy'
            if gen[i]['type'] == 'CoolingTower':
                c_red['ab'][i] = 'cw'
    return c_red


def half_loop_tank(volume,den,flow,T_old,pump,T_inlet,dt):
    if flow>0:
        Cp_water = 4186  #J/kg*K
        e = 2.718281828459 # natural logarithm
        M_tank = volume*den #Mass of fluid in tank
        A = (flow*Cp_water*T_inlet+pump)/(flow*Cp_water) #Watts / J/s*K
        T_new = (T_old - A)*e**(-flow*Cp_water*dt/(M_tank*Cp_water)) + A 
        # T_average = (T_old - A)*e**(-flow*Cp_water*dt/(2*M_tank*Cp_water)) + A #temperature halfway through time period
        T_average = M_tank/(flow*dt)*(T_old - A)*(1-e**(-flow*dt/M_tank)) + A #Reference manual page 468
    else:
        T_new = T_old+0
        T_average = T_old+0
    return T_new,T_average


import datetime as dt


# Types of data sets.
DATASET_TYPES = ('nodedata', 'weather')

# HDF5 table descriptions.
DATASET_DESCR = dict(
    nodedata=dict(
        timestamp=dict(col='Time64Col', pos=0),
        e=dict(col='Float64Col', pos=1),
    ),
    weather=dict(
        timestamp=dict(col='Float32Col', pos=0),
        glo_horz_irr=dict(col='Float32Col', pos=1),
        dir_norm_irr=dict(col='Float32Col', pos=2),
        dif_horz_irr=dict(col='Float32Col', pos=3),
        t_dryb=dict(col='Float32Col', pos=4),
        t_dewp=dict(col='Float32Col', pos=5),
        rh=dict(col='Float32Col', pos=6),
        pres_pa=dict(col='Float32Col', pos=7),
        wdir=dict(col='Float32Col', pos=8),
        wspd=dict(col='Float32Col', pos=9),
        tot_cld=dict(col='Float32Col', pos=10),
        opq_cld=dict(col='Float32Col', pos=11),
    ),
)

# Conversion from shape codes to dimension info field names.
# These provide the mapping between given dimension info field names and
# the letters used in column shape specifications.
#
# ASSUMPTION:
# column_realization(), in util/hdf5.py, assumes that every shape code
# is a single letter.  Be careful when adding new ones to either follow
# the current convention or update it.
SHAPECODE_MAP = dict(
    g='generator',
    n='line',  # 'l' looks too much like '1' on its own.
    b='building',
    z='zone',
    f='fluid_loop',
    w='hydro',
    h='horizon',
    s='solution',
)

# Default HDF5 table name.
DEFAULT_TABLENAME = 'hourly'
# Mapping from timestamp resolution to table names.
TIMESTAMP_RESOLUTION_TABLENAMES = {
    dt.timedelta(hours=1): 'hourly',
    dt.timedelta(days=1): 'daily',
}

# Data scrubbing.
SCRUB_LEVEL_ATTR = 'scrub_level'
SCRUB_VERSION_ATTR = 'scrub_version'
SCRUB_LEVELS = {
    'v1': {
        0: 'raw',
        1: 'missing filled',
    },
    'v2': {
        0: 'raw',
        1: 'zeros trimmed',
        2: 'DST removed',
        3: 'missing filled',
        4: 'partitioned',
    }
}
MAX_SCRUB_VERSION = max(int(v[1:]) for v in SCRUB_LEVELS.keys())
NEWEST_SCRUB_LEVELS = SCRUB_LEVELS[f"v{MAX_SCRUB_VERSION}"]



"""Defines variables relating to chosen precision values.
"""


# Datetime rounding precision.
# Example: If precision is 'second':
#   1. 1/2 sec will be added.
#   2. However many many milliseconds (the next highest precision level)
#       the datetime now has will be subtracted from it.  I.e. the
#       datetime will be "floored" in terms of milliseconds.

# Precision used in rounding datetimes.
DATETIME_ROUNDING_PRECISION = 'second'

# timedelta field to be used in "flooring" to a rounded datetime.
DATETIME_ROUNDING_TIMEDELTA_FIELD = 'microsecond'



from eagers.building.basic.psychometric import psychometric
from eagers.building.basic.two_way_interp import two_way_interp
from eagers.building.basic.eval_curve import eval_curve

import numpy as np

def cooling_coil(building,name,air_in,T_outside,T_treat):
    ## Compute the cooling and electric load for DX cooling coils (local AC units not on a central water loop)
    coil = building.coils_cooling[name]
    curve = building.curves
    air_density = 1.204  #kg/m^3  #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    air_out = {}
    for k in air_in:
        air_out[k] = air_in[k]+0
    spec_heat_supply = 1006 + 1860*air_in['w']  #J/kg*K
    sensible_load = air_in['m_dot']*spec_heat_supply*(air_in['Tdb'] - T_treat)
    if sensible_load>10:
        if coil['type']=='Cooling:DX:SingleSpeed' or coil['type']=='Cooling:DX:TwoSpeed': #page 805
            try: 
                c_n = building.hvac['components']['name'].index(name)
                al = building.hvac['components']['loop'][c_n] #air loop
                bf_rated = bypass_factor_rated(building.hvac['design']['Tsupply_c'][al],building.hvac['design']['w_supply_c'][al],building.hvac['design']['all_outdoor_flow_cooling'][al])
            except ValueError:
                al =[]
                for uhc in building.unitary_heat_cool:
                    if ('cool_coil' in building.unitary_heat_cool[uhc] and building.unitary_heat_cool[uhc]['cool_coil']==name):
                        al = building.unitary_heat_cool[uhc]['loop']
                        bf_rated = bypass_factor_rated(building.hvac['design']['Tsupply_c'][al],building.hvac['design']['w_supply_c'][al],building.hvac['design']['all_outdoor_flow_cooling'][al])
                        break
                if len(al)==0:
                    for us in building.unitary_sys:
                        if ('cool_coil' in building.unitary_sys[us] and building.unitary_sys[us]['cool_coil']==name):
                            z = building.unitary_sys[us]['zone']
                            sp = building.setpoints['zone'].index(z)
                            bf_rated = bypass_factor_rated(building.setpoints['cooling_T_des'][sp],building.setpoints['cooling_w_des'][sp],False)
                            break
                    
            coil['rated_A0'] = float(-np.log(bf_rated))*coil['rated_air_flow']*air_density
            
            bypass_factor = float(np.exp(-coil['rated_A0']/air_in['m_dot']))
            input = [psychometric(air_in,'Twb'),T_outside]  #needs to be outdoor wetbulb temperature if it is an evaporative cooled condenser
            
            TotCapTempMod= eval_curve(curve,coil['capacity_v_temperature_curve'],input)
            EIRTempMod = eval_curve(curve,coil['energy_input_v_temperature_curve'],input)

            input2 = air_in['m_dot']/(coil['rated_air_flow']*air_density)  #normalized flow
            TotCapFlowMod = eval_curve(curve,coil['capacity_v_flow_curve'],[input2]) 
            EIRFlowMod= eval_curve(curve,coil['energy_input_v_flow_curve'],[input2])
            EIR = (1/coil['rated_COP'])*EIRTempMod*EIRFlowMod #energy input ratio
            Q_total = coil['rated_capacity']*TotCapTempMod*TotCapFlowMod
            PLR = min(1,sensible_load/Q_total)
            Q_max = Q_total
            Speed_ratio = 1
            if coil['type']=='Cooling:DX:TwoSpeed':
                TotCapTempMod_2 = eval_curve(curve,coil['capacity_v_temperature_curve2'],input)
                EIRTempMod_2 = eval_curve(curve,coil['energy_input_v_temperature_curve2'],input)

                Q_total2 = coil['rated_capacity2']*TotCapTempMod_2*TotCapFlowMod
                EIR_2 = (1/coil['rated_COP2'])*EIRTempMod_2*EIRFlowMod #energy input ratio

                Speed_ratio = max(0,(sensible_load - Q_total2)/(Q_total - Q_total2))
                if Speed_ratio<0:
                    PLR = min(1,sensible_load/Q_total2)
                    Q_max = Q_total2
            error = 100
            old_error = 100
            while abs(error)>1e-4:
                air_out = coil_air_out(air_in,Q_max,bypass_factor,PLR)
                air_out2 = coil_air_out(air_in,Q_max,bypass_factor,PLR+1e-3)
                dT_dPLR = (air_out['Tdb'] - air_out2['Tdb'])/1e-3
                if dT_dPLR<0:
                    dT_dPLR = .05
                error = (air_out['Tdb'] - T_treat)
                if (PLR == 1 and error>0) or (error<0 and PLR == 0):
                    error = 0
                if error>old_error:
                    PLR = max(0,min(1,PLR + error/(4*dT_dPLR)))
                else:
                    PLR = max(0,min(1,PLR + error/dT_dPLR))
                old_error = error
            if coil['type']=='Cooling:DX:TwoSpeed' and Speed_ratio>0:
                revised_sensible_load = PLR*Q_total
                Speed_ratio = max(0,(revised_sensible_load - Q_total2)/(Q_total - Q_total2))
                Power = Speed_ratio*Q_total*EIR + (1-Speed_ratio)*Q_total2*EIR_2
                heat_rejected = Speed_ratio*Q_total*(1+EIR) + (1-Speed_ratio)*Q_total2*(1+EIR_2)
            else:
                PartLoadFrac = max(.7,eval_curve(curve,coil['part_load_curve'],[PLR]))
                RTF = PLR/PartLoadFrac #runtime fraction
                if coil['type']=='Cooling:DX:TwoSpeed':
                    Power = Q_total2*EIR_2*RTF
                    heat_rejected = Q_total2*(1+EIR_2)
                else:
                    Power = Q_total*EIR*RTF
                    heat_rejected = Q_total*(1+EIR)
    else:
        Power = 0
    return air_out,Power

def coil_air_out(air_in,Q_total,bypass_factor,PLR):
    ## Determine exit air conditions based on operating status of the coil

    #find apparatus dewpoint
    h_ADP = air_in['h'] - (Q_total/air_in['m_dot'])/(1-bypass_factor) # J/kg
    T_search = [2*i+2 for i in range(19)]
    w_ADP = two_way_interp(h_ADP,psychometric({'Twb':T_search},'h'),psychometric({'Twb':T_search},'w'))
    air_out = {}
    air_out['m_dot'] = air_in['m_dot']
    air_out['h'] = air_in['h'] - Q_total*PLR/air_out['m_dot'] #J/kg
    if air_in['w']<w_ADP:  #dry coil
        sensible_heat_ratio = 1
        air_out['w'] = air_in['w']
    else:
        sensible_heat_ratio = min(1,(psychometric({'Tdb':air_in['Tdb'],'w':w_ADP},'h')-h_ADP)/(air_in['h'] - h_ADP))
        h_Tin_w_out = air_in['h'] - (1-sensible_heat_ratio)*(air_in['h'] - air_out['h'])
        air_out['w'] = psychometric({'Tdb':air_in['Tdb'],'h':h_Tin_w_out},'w') 
    spec_heat = 1006 + 1860*(air_in['w']+air_out['w'])/2  #J/kg*K
    air_out['Tdb'] = air_in['Tdb'] - sensible_heat_ratio*Q_total*PLR/(air_out['m_dot']*spec_heat)
    return air_out

def bypass_factor_rated(T_supply,w_supply,all_outdoor):
    ## determine bypass factor based on rated dew point
    if all_outdoor:
        T_mix = 26.7
        w_mix = 0.01115  #T_wb = 19.5, RH = 50.7 #
    else:
        T_mix = 35
        w_mix = 0.014  #T_wb = 23.9, RH = 40 #
    h_mix = psychometric({'Tdb':T_mix,'w':w_mix},'h')
    h_sup = psychometric({'Tdb':T_supply,'w':w_supply},'h')

    rated_slope = (w_mix - w_supply)/(T_mix - T_supply)
     # #find apparatus dewpoint
    T_search = [5+i for i in range(round(T_supply+.5))]
    W_s = psychometric({'Twb':T_search},'w')
    W_coil_linear_extrap = [rated_slope*(T_search[i]-T_supply) + w_supply for i in range(len(T_search))]
    for i in range(len(W_s)):
        if W_s[i]>=W_coil_linear_extrap[i]:
            break
    apparatus_dewpoint = T_search[i]

    h_ADP = psychometric({'Tdb':apparatus_dewpoint,'w':W_s[i]},'h')
    bf = (h_sup - h_ADP)/(h_mix-h_ADP)
    return bf




def water_equipment(we,profile,T_mains,T_tank):    
    if we['cold_supply_temperature_schedule'] == None:
        T_cold = T_mains
    else:
        T_cold = profile[we['cold_supply_temperature_schedule']]
    if T_tank==None:
        T_hot =  profile[we['hot_supply_temperature_schedule']]
    else:
        T_hot = T_tank 
    if we['target_temperature_schedule']==None:
        T_target = T_mains
    else:
        T_target = min(T_hot,profile[we['target_temperature_schedule']])
    den_water = 997  #kg/m^3
    m_mix = we['peak_flow']*profile[we['flow_schedule']]*den_water
    m_hot = m_mix*(T_target- T_cold)/(T_hot - T_cold)
    m_cold = m_mix-m_hot
    flow = m_hot
    T_return = T_cold
    return flow,T_return,T_target,T_hot,T_cold


from eagers.setup.load_network import load_network
from eagers.setup.update_qpform_all import update_qpform_all
from eagers.setup.add_buildings import add_buildings
from eagers.basic.build_time_vector import build_time_vector
from eagers.basic.check_ac_dc import check_ac_dc
from eagers.setup.load_matrices import load_matrices

def preload(plant,test_data,options,online=None):
    preload= {}
    time = build_time_vector(options['start_date'], options, to_timedelta=True)
    dt = [(time[i+1] - time[i]).total_seconds()/3600 for i in range(len(time)-1)]
    subnet = load_network(plant['network'], plant['generator'], plant['building'])
    gen_qp_form =  update_qpform_all(plant['generator'], subnet, test_data, 1)
    building_qp_form = add_buildings(plant['building'], subnet, gen_qp_form)
    gen_qp_form = check_ac_dc(gen_qp_form, building_qp_form, test_data)
    b_names = [plant['building'][i].name for i in range(len(plant['building']))]
    op_mat_a = load_matrices(gen_qp_form, b_names, plant['fluid_loop'], plant['market'], subnet, options, 'a', dt)
    op_mat_b = load_matrices(gen_qp_form, b_names, plant['fluid_loop'], plant['market'], subnet, options, 'b', dt)
    one_step = load_matrices(gen_qp_form, b_names, plant['fluid_loop'], plant['market'], subnet, options, 'b', None)
    if online is None and options['method'] == 'Control':
        a = {}
        a['horizon'] = options['resolution'] #the horizon is the resolution
        a['resolution'] = options['t_opt']/3600 #the resolution is the frequency of Topt
        a['tspacing'] = 'constant'
        
        online_time = build_time_vector(options['start_date'], a, to_timedelta=True), ## set up dt vector of time interval length
        ns = len(online_time)
        pre_time = [0]
        pre_time.extend([online_time[i+1] for i in range(ns-1)])
        dt2 = [online_time[i] - pre_time[i] for i in range(ns)]
        online = [[] for i in range(ns)]
        for t in range(ns):
            online[t] = load_matrices(gen_qp_form, b_names, plant['fluid_loop'], plant['market'], subnet, options, 'b', [dt2[i] for i in range(t, ns)]) #build the matrix for the onlineOptimLoop using FitB:
    else:
        online = []

    preload['building_qp_form'] = building_qp_form
    preload['gen_qp_form'] = gen_qp_form
    preload['num_steps'] = 0
    preload['one_step'] = one_step
    preload['online'] = online
    preload['op_mat_a'] = op_mat_a
    preload['op_mat_b'] = op_mat_b
    preload['subnet'] = subnet
    return preload